﻿using System;
using System.Text;

namespace OrientingMap
{
    public class QuadKey
    {
        long value = long.MinValue;
        int longitude = int.MinValue;
        int latitude = int.MinValue;
        QuadKey parent;
        QuadKey[] siblings;
        QuadKey[] children;
        string strValue;

        public QuadKey(long value, int level)
        {
            this.value = value;
            this.Level = level;
        }

        public QuadKey(int longitude, int latitude, int level)
        {
            this.longitude = longitude;
            this.latitude = latitude;
            this.Level = level;
        }

        public static string UriTemplate { set; get; }

        public long Value 
        {
            get
            {
                if (this.value == long.MinValue)
                {
                    this.value = 0;
                    int mask = 1 << (this.Level - 1);

                    for (int i = 0; i < this.Level; i++)
                    {
                        value <<= 2;

                        if ((this.longitude & mask) != 0)
                            this.value |= 1;

                        if ((this.latitude & mask) != 0)
                            this.value |= 2;

                        mask >>= 1;
                    }
                }
                return this.value;
            }
        }

        public int Longitude 
        {
            get
            {
                if (this.longitude == int.MinValue)
                {
                    CalculateLongitudeLatitudeFromQuadKey();
                }
                return this.longitude;
            }
        }

        public int Latitude 
        {
            get
            {
                if (this.latitude == int.MinValue)
                {
                    CalculateLongitudeLatitudeFromQuadKey();
                }
                return this.latitude;
            }
        }

        public int Level { private set; get; }

        public int Quadrant
        {
            get { return (int)(this.Value & 3); }
        }

        public QuadKey Parent
        {
            get
            {
                if (this.parent == null)
                {
                    this.parent = new QuadKey(this.Value >> 2, this.Level - 1);
                }
                return this.parent;
            }
        }

        // Not used
        public QuadKey[] Siblings
        {
            get
            {
                if (this.siblings == null)
                {
                    this.siblings = new QuadKey[4];
                    long quadKey0 = this.Value & ~3;

                    for (uint i = 0; i < 4; i++)
                        this.siblings[i] = new QuadKey(quadKey0 | i, this.Level);
                }
                return this.siblings;
            }
        }

        public QuadKey[] Children
        {
            get
            {
                if (this.children == null)
                {
                    this.children = new QuadKey[4];
                    long quadKey0 = this.Value << 2;

                    for (uint i = 0; i < 4; i++)
                        this.children[i] = new QuadKey(quadKey0 | i, this.Level + 1);
                }
                return this.children;
            }
        }

        public QuadKey GetQuadKeyFromRowColumnOffsets(int col, int row)
        {
            int maxValue = (1 << this.Level) - 1;

            if (this.Longitude + col < 0 ||
                this.Longitude + col > maxValue ||
                this.Latitude + row < 0 ||
                this.Latitude + row > maxValue)
            {
                return null;
            }

            return new QuadKey(this.Longitude + col, this.Latitude + row, this.Level);
        }

        void CalculateLongitudeLatitudeFromQuadKey()
        {
            this.longitude = 0;
            this.latitude = 0;
            long mask = 1L << (2 * this.Level - 1);

            for (int i = 0; i < this.Level; i++)
            {
                this.longitude <<= 1;
                this.latitude <<= 1;

                if ((this.value & mask) != 0)
                    this.latitude |= 1;

                mask >>= 1;

                if ((this.value & mask) != 0)
                    this.longitude |= 1;

                mask >>= 1;
            }
        }
        
        public Uri ToUri()
        {
            return new Uri(UriTemplate.Replace("{quadkey}", this.ToString()));
        }

        public bool Equals(QuadKey that)
        {
            return this.Value == that.Value && 
                   this.Level == that.Level;
        }

        public override bool Equals(object obj)
        {
            if (!(obj is QuadKey))
                return false;

            return Equals(obj as QuadKey);
        }

        public override int GetHashCode()
        {
            return (int)this.Value;
        }

        public override string ToString()
        {
            if (strValue == null)
            {
                StringBuilder strBuilder = new StringBuilder();
                long quadKey = this.Value;

                for (int i = 0; i < this.Level; i++)
                {
                    strBuilder.Insert(0, (quadKey & 3).ToString());
                    quadKey >>= 2;
                }
                strValue = strBuilder.ToString();
            }
            return strValue;
        }
    }
}
